import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import read_cifar as rc

# learn_once_mse
def learn_once_mse(w1, b1, w2, b2, data, targets, learning_rate):
    """
    Update the weights and biases of the network for one gradient descent step using Mean Squared Error (MSE) loss.

    Parameters:
    - w1: Weight matrix of the first layer (shape: d_in x d_h).
    - b1: Bias vector of the first layer (shape: 1 x d_h).
    - w2: Weight matrix of the second layer (shape: d_h x d_out).
    - b2: Bias vector of the second layer (shape: 1 x d_out).
    - data: Input data matrix (shape: batch_size x d_in).
    - targets: Target output matrix (shape: batch_size x d_out).
    - learning_rate: Learning rate for gradient descent.

    Returns:
    - W1: Updated weight matrix of the first layer.
    - b1: Updated bias vector of the first layer.
    - w2: Updated weight matrix of the second layer.
    - b2: Updated bias vector of the second layer.
    - loss: Mean Squared Error (MSE) loss for monitoring.
    """
    a0 = data
    # Forward pass
    z1 = np.matmul(a0, w1) + b1  # Calculate the weighted sum for the hidden layer
    a1 = 1 / (1 + np.exp(-z1))  # Apply the sigmoid activation function to hidden layer
    z2 = np.matmul(a1, w2) + b2  # Calculate the weighted sum for the output layer
    a2 = np.exp(z2) / np.sum(np.exp(z2), axis=1, keepdims=True)  # Apply the softmax activation to the output layer
    predictions = a2  # The network's predictions

    n = data.shape[0]  # Number of samples (batch size)

    # Backpropagation
    e2 = predictions - targets  # Compute the error in the output layer

    dw2 = e2 * a2 * (1 - a2) / n  # Gradient for w2
    update_w2 = np.dot(a1.T, dw2) / n # Update for w2
    update_b2 = (1/a1.shape[1])*dw2.sum(axis=0, keepdims=True) # Update for b2

    e1 = np.dot(e2, w2.T) # Compute the error in the hidden layer
    dw1 = e1 * a1 * (1 - a1)  # Gradient for w1
    update_b1 = (1/data.shape[1])*dw1.sum(axis=0, keepdims=True)  # Update for b1
    update_w1 = np.dot(data.T, dw1) / n # Update for w2

    # Gradient descent    
    w2 = w2 - learning_rate * update_w2
    b2 = b2 - learning_rate * update_b2
    w1 = w1 - learning_rate * update_w1
    b1 = b1 - learning_rate * update_b1

    # Calculate the Mean Squared Error (MSE) loss
    loss = compute_error(predictions, targets, loss_type = 'MSE')  
    
    return w1, b1, w2, b2, loss  

#One Hot Function :
def one_hot(targets):
    """
    one_hot_encode takes an arrayy of target values and returns the corresponding one-hot encoded matrix.
    
    Parameters:
    - targets: An arrayy of target values.
    
    Returns:
    - one_hot_matrix: A one-hot encoded matrix where each row corresponds to a target value.
    """
    num_classes = np.unique(targets).shape[0]  # Determine the number of unique classes in the target arrayy
    num_samples = targets.shape[0]  # Get the number of samples in the target arrayy

    one_hot_matrix = np.zeros((num_samples, num_classes))  # Initialize a matrix of zeros

    for i in range(num_samples):
        target_class = targets[i]
        one_hot_matrix[i, target_class] = 1  # Set the corresponding class index to 1

    return one_hot_matrix


# The function learn_once_mse:
def learn_once_cross_entropy(w1, b1, w2, b2, data, targets, learning_rate):
    """
    Perform one gradient descent step using binary cross-entropy loss.

    Parameters:
    - W1, b1, W2, b2: Weights and biases of the network.
    - data: Input data matrix of shape (batch_size x d_in).
    - targets: Target output matrix of shape (batch_size x d_out).
    - learning_rate: Learning rate for gradient descent.

    Returns:
    - Updated weights and biases (W1, b1, W2, b2) of the network.
    - Loss value for monitoring.
    """
    # Forward pass
    z1 = np.matmul(data, w1) + b1
    a1 = 1 / (1 + np.exp(-z1)) 
    z2 = np.matmul(a1, w2) + b2
    a2 = np.exp(z2) / np.sum(np.exp(z2), axis=1, keepdims=True)  

    predictions = a2

    one_hot_matrix = one_hot(targets)

    n = data.shape[0] 

    # Backpropagation
    e2 = predictions - one_hot_matrix
    dw2 = e2 * a2 * (1 - a2) / n  
    update_w2 = np.dot(a1.T, dw2) / n
    update_b2 = (1/a1.shape[1])*dw2.sum(axis=0, keepdims=True) 

    e1 = np.dot(e2, w2.T)
    dw1 = e1 * a1 * (1 - a1)  
    update_b1 = (1/data.shape[1])*dw1.sum(axis=0, keepdims=True)
    update_w1 = np.dot(data.T, dw1) / n

    # Gradient descent
    w2 = w2 - learning_rate * update_w2
    b2 = b2 - learning_rate * update_b2
    w1 = w1 - learning_rate * update_w1
    b1 = b1 - learning_rate * update_b1
    
    # Calculate binary cross-entropy loss
    loss = compute_error(predictions, one_hot_matrix, loss_type = 'binary cross-entropy') 
    
    # Calculate the accuray for a single batch
    batch_accuracy = accuracy(predictions, one_hot_matrix) 

    return w1, b1, w2, b2, loss, batch_accuracy

#  The function train_mlp:
def train_mlp(w1, b1, w2, b2, data_train, labels_train, learning_rate, num_epoch):
    """ 
    Perform training steps for a specified number of epochs.

    Parameters:
    - W1, b1, W2, b2: Weights and biases of the network.
    - data: Input data matrix of shape (batch_size x d_in).
    - targets: Target output matrix of shape (batch_size x d_out).
    - learning_rate: Learning rate for gradient descent.
    - num_epochs: Number of training epochs.
    - metrics: Specifies the loss metric (default is Binary Cross Entropy).

    Returns:
    - Updated weights and biases (W1, b1, W2, b2) of the network.
    - List of training accuracies across epochs as a list of floats.
    """
    train_accuracies = []  # To store training accuracies across epochs

    # Iterate through the specified number of epochs
    for epoch in range(num_epoch):

        # Call the 'learn_once_cross_entropy' function to update weights, calculate loss, and obtain batch accuracy
        w1, b1, w2, b2, loss, batch_accuracy = learn_once_cross_entropy(w1, b1, w2, b2, data_train, labels_train, learning_rate)

        # Append the batch accuracy to the 'train_accuracies' list for tracking progress
        train_accuracies.append(batch_accuracy)

        # Print the current epoch's progress
        print("Epoch {}/{}".format(epoch+1, num_epoch))
        print("[=======] Train_Accuracies : {}".format(round(batch_accuracy, 5)))

    return w1, b1, w2, b2, train_accuracies

# The function test_mlp:
def test_mlp(w1,b1,w2,b2,data_test,labels_test):
    """
     Evaluate the network's performance on the test set.

    Parameters:
    - W1, b1, W2, b2: Weights and biases of the network.
    - data_test: Test data matrix of shape (batch_size x d_in).
    - labels_test: True labels for the test data.

    Returns:
    - test_accuracy: The testing accuracy as a float.
    """
    z1 = np.matmul(data_test, w1) + b1
    a1 = 1 / (1 + np.exp(-z1))  
    z2 = np.matmul(a1, w2) + b2
    a2 = np.exp(z2) / np.sum(np.exp(z2), axis=1, keepdims=True)  

    # Compute the testing accuracy using the 'accuracy' function
    test_accuracy = accuracy(a2, labels_test)

    return test_accuracy 


# The function run_mlp_training:
def run_mlp_training(X_train, labels_train, data_test, labels_test, d_h, learning_rate, num_epoch):
    """
    Train an MLP classifier and evaluate its performance.

    Parameters:
    - X_train: Training data matrix of shape (batch_size x input_dimension).
    - labels_train: True labels for the training data.
    - data_test: Test data matrix of shape (batch_size x input_dimension).
    - labels_test: True labels for the test data.
    - num_hidden_units: Number of neurons in the hidden layer.
    - learning_rate: The learning rate for gradient descent.
    - num_epochs: The number of training epochs.

    Returns:
    - train_accuracies: List of training accuracies across epochs.
    - test_accuracy: The final testing accuracy.
    """
    
    d_in = X_train.shape[1]   # Input dimension
    d_out = 10  # Output dimension: 10 classes

    np.random.seed(10)  # Set a random seed for reproducibility

    # Initialize weights and biases for the neural network
    w1 = 2 * np.random.rand(d_in, d_h) - 1  # First layer weights
    b1 = np.zeros((1, d_h))  # First layer biases
    w2 = 2 * np.random.rand(d_h, d_out) - 1  # Second layer weights
    b2 = np.zeros((1, d_out))  # Second layer biases

    # Train the MLP using the provided training data and parameters
    w1, b1, w2, b2, train_accuracies = train_mlp(w1, b1, w2, b2, X_train, labels_train, learning_rate, num_epoch)

    # Test the trained MLP on the testing data and compute the test accuracy
    test_accuracy = test_mlp(w1, b1, w2, b2, data_test, one_hot(labels_test))

    # Print the test set accuracy
    print("test accuracy:", test_accuracy)

    return train_accuracies, test_accuracy

# Plot of the evolution of learning accuracy across learning epochs:
def plot_ANN(data_train, labels_train, data_test, labels_test):
    """
    Plot the variation of accuracy in terms of the number of epochs.

    Parameters:
    - X_train: Training data matrix.
    - y_train: True labels for the training data.
    - X_test: Test data matrix.
    - y_test: True labels for the test data.
    """
    # Train the MLP and obtain training accuracies and test accuracy
    train_accuracies, test_accuracy = run_mlp_training(data_train, labels_train, data_test, labels_test, 64, 0.1, 100)

    # Create a DataFrame from the accuracy values
    df = pd.DataFrame({'Epoch': range(1, len(train_accuracies) + 1), 'Accuracy': train_accuracies})

    # Create a line plot using Matplotlib
    plt.figure(figsize=(10, 6))
    plt.plot(df['Epoch'], df['Accuracy'], 'b')

    # Add labels and title to the plot
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.title('The Variation of Accuracy')

    # Save the plot as an image file
    plt.savefig("Results/mlp.png")

# Define accuracy function
def accuracy(y_pred, y_true):
    """
    calculate_accuracy: Compute the accuracy of the model.

    Parameters:
    - predictions: Predicted values.
    - actual_values: Ground truth observations.

    Returns:
    - Accuracy as a float.
    """
    accuracy = (y_pred.argmax(axis=1) == y_true.argmax(axis=1)).mean()
    return accuracy

def compute_error(predictions, targets, loss_type):
    # Calculate the loss based on the specified loss type
    
    if loss_type == 'MSE':  # Mean Squared Error loss
        loss = np.mean(np.square(predictions - targets))
    elif loss_type == 'binary cross-entropy':  # Binary Cross-Entropy loss
        n = targets.shape[0]
        loss = -(1/n)*np.mean((np.dot(targets.T,np.log(predictions+ 1e-7)) + np.dot((1 - targets).T,np.log((1 - predictions+ 1e-7)))))
    else:
        raise ValueError("Unsupported loss type. Use 'MSE' or 'binary cross-entropy'.")
    
    return loss

# The following code block is executed only if the script is run as the main program

if _name_ == "_main_":

    # Read the CIFAR-10 dataset from the specified path
    X, y = rc.read_cifar('data\cifar-10-batches-py')
    
    # Split the dataset into training and testing sets
    X_train, y_train, X_test, y_test = rc.split_dataset(X, y, split=0.9)
    
    # Plot the evolution of learning accuracy across learning epochs using the 'plot_ANN' function
    plot_ANN(X_train, y_train, X_test, y_test)