import numpy as np
import os
import pickle

# Function read_cifar_batch:
'''
Arguments:
A single batch's path as a string.

Returns:
A data matrix of size (batch_size, data_size)
A labels vector of size (batch_size)
'''

def read_cifar_batch(batch_path):
    # Opens the batch file and loads the data
    with open(batch_path, 'rb') as f:
        batch_dict = pickle.load(f, encoding='bytes')
    # Converts the data to float32
    data = batch_dict[b'data'].astype(np.float32)
    # Converts the labels to int64
    labels = np.array(batch_dict[b'labels'], dtype=np.int64)
    return data, labels

# Function read_cifar:
'''
Reads the directory path containing all the batches (including test_batch).

Arguments:
The directory path containing all six batches (five data_batch and one test_batch) as a string.

Returns:
- A data matrix of size (total_batch_size, data_size).
- A labels vector of size (total_batch_size).
'''
def read_cifar(folder):
    # List of batch file names
    batch_files = ["data_batch_1", "data_batch_2", "data_batch_3", "data_batch_4", "data_batch_5", "test_batch"]
    data_list, labels_list = []

    # Loop over batch files
    for batch_file in batch_files:
        path = os.path.join(folder, batch_file)
        # Calls read_cifar_batch to read each batch
        data, labels = read_cifar_batch(path)
        data_list.append(data)
        labels_list.append(labels)

    # Combines data from all batches
    data = np.vstack(data_list)
    # Combines labels from all batches
    labels = np.hstack(labels_list)

    return data, labels

# Function to split the data into training and testing sets:
'''
Arguments:
- data and labels, two arrays of the same size in the first dimension.
- split, a floating-point number between 0 and 1, determining the split ratio of the training set compared to the test set.

Returns:
- data_train: the training data.
- labels_train: the corresponding labels.
- data_test: the testing data.
- labels_test: the corresponding labels.
'''
def split_dataset(data, labels, split):
    # Check that the split ratio is valid
    assert 0 < split < 1
    n = len(data)
    # Shuffle the data indices
    indices = np.random.permutation(n)
    split_index = int(split * n)
    # Separate the indices for the training and test sets
    train_indices = indices[:split_index]
    test_indices = indices[split_index:]
    # Split the data and labels into training and test sets
    data_train = data[train_indices]
    labels_train = labels[train_indices]
    data_test = data[test_indices]
    labels_test = labels[test_indices]
    return data_train, labels_train, data_test, labels_test